#include <stdio.h>
#include <stdint.h>     /* uint8_t */
#include <stdlib.h>     /* calloc et al.*/
#include <string.h>     /* memset */

#include <avcodec.h>
#include <avformat.h>

/*
 gcc -O2 -o ffbench ffbench.c  \
 `pkg-config --cflags --libs libavcodec libavformat libavutil libswscale` -DHAVE_SWSCALE \ 
 -I/usr/include/ffmpeg/libavutil/ -I/usr/include/ffmpeg/libavformat/ -I/usr/include/ffmpeg/libavcodec/ -I/usr/include/ffmpeg/libswscale 

*/

#ifdef HAVE_SWSCALE
#include <avutil.h>
#include <swscale.h>
#endif

int want_verbose=0;
int want_quiet=0;

enum {  SEEK_ANY, ///< directly seek to givenvideo frame
        SEEK_KEY, ///< seek to next keyframe after given frame.
        SEEK_CONTINUOUS }; ///< seek to keframe before this frame and advance to current frame.

/* ffmpeg source */
typedef struct {
  /* file specific decoder settings */
  int   want_ignstart; //< set before calling ff_open_movie()
  int   want_genpts;
  int   seekflags;
  /* Video File Info */
  int   out_width;
  int   out_height;
  double duration;
  double framerate;
  double file_frame_offset;
  long   frames;
  char  *current_file;
  /* helper variables */
  double tpf;
  int64_t avprev;
  /* */
  uint8_t *buffer;
  int   videoStream;
  int   render_fmt;  //< pFrame/buffer output format (RGB24)
  /* ffmpeg internals*/
  AVPacket          packet; 
  AVFormatContext   *pFormatCtx;
  AVCodecContext    *pCodecCtx;
  AVFrame           *pFrame;
  AVFrame           *pFrameFMT;
#ifdef HAVE_SWSCALE
  struct SwsContext *pSWSCtx;
#endif
 
} ffst;

#include <time.h>
#include <math.h>
#include <getopt.h>
#include <sys/time.h>
#include <unistd.h>

/* Option flags and variables */
extern int want_quiet;
extern int want_verbose;

const AVRational c1_Q = { 1, 1 };

//--------------------------------------------
// Manage video file
//--------------------------------------------

void ff_getbuffersize(void *ptr, size_t *s) {
  ffst *ff=(ffst*)ptr;
  *s=(avpicture_get_size(ff->render_fmt, ff->out_width, ff->out_height));
}

void ff_init_moviebuffer(void *ptr) {
  size_t numBytes;
  ffst *ff=(ffst*)ptr;

  ff->out_width  = ff->pCodecCtx->width;
  ff->out_height = ff->pCodecCtx->height;

  if (ff->buffer) free(ff->buffer);
  ff_getbuffersize(ff,&numBytes);
  ff->buffer=(uint8_t *) calloc(numBytes,sizeof(uint8_t));
  if (!ff->buffer) { 
    fprintf(stderr, "out of memory -tying to alloc %u bytes\n",numBytes); 
    exit(1);
  }
  if (!ff->pFrameFMT) { // Assign appropriate parts of buffer to image planes in pFrameFMT
    fprintf(stderr, "could not initialize output frame/scaling.\n"); 
    exit(1); // TODO: don't die
  }
  avpicture_fill((AVPicture *)ff->pFrameFMT, ff->buffer, ff->render_fmt, ff->out_width, ff->out_height);
}

void ff_set_framerate(ffst *ff) {
  AVStream *av_stream;
// At LIBAVFORMAT_BUILD==4624 r_frame_rate becomes an AVRational. Before it was an int.
  av_stream = ff->pFormatCtx->streams[ff->videoStream];
#if LIBAVFORMAT_BUILD <= 4616
  ff->framerate = (double) av_stream->codec.frame_rate / (double) av_stream->codec.frame_rate_base;
#elif LIBAVFORMAT_BUILD <= 4623 // I'm not sure that this is correct:
  ff->framerate = (double) av_stream->r_frame_rate / (double) av_stream->r_frame_rate_base;
#else
  if(av_stream->r_frame_rate.den && av_stream->r_frame_rate.num) {
    ff->framerate = av_q2d(av_stream->r_frame_rate);
    if ((ff->framerate < 4 || ff->framerate > 100 ) && (av_stream->time_base.num && av_stream->time_base.den))
      ff->framerate = 1.0/av_q2d(av_stream->time_base);
  } else ff->framerate = 1.0/av_q2d(av_stream->time_base);
#endif
}

void ff_initialize (void) {
  printf("FFMPEG: registering codecs.\n");
  av_register_all();
  avcodec_init();
  avcodec_register_all();
  if(!want_verbose) av_log_set_level(AV_LOG_QUIET);
}

int ff_close_movie(void *ptr) {
  ffst *ff=(ffst*)ptr;
  if(ff->current_file) free(ff->current_file);
  ff->current_file=NULL;

  if (!ff->pFrameFMT) return(-1);

#ifdef HAVE_SWSCALE
  //if (ff->pSWSCtx) sws_freeContext(ff->pSWSCtx); 
#endif

  //if (ff->buffer) free(ff->buffer); // done in pFrameFMT?
  if (ff->pFrameFMT) av_free(ff->pFrameFMT);
  if (ff->pFrame) av_free(ff->pFrame);
  ff->buffer=NULL;ff->pFrameFMT=ff->pFrame=NULL;
  avcodec_close(ff->pCodecCtx);
  av_close_input_file(ff->pFormatCtx);
  return (0);
}

int ff_open_movie(void *ptr, const char *file_name) {
  int i;
  AVCodec *pCodec;
  ffst *ff = (ffst*) ptr;

  if (ff->pFrameFMT) {
    if (ff->current_file && !strcmp(file_name, ff->current_file)) return(0);
    /* close currently open movie */
    if (!want_quiet)
    fprintf(stderr,"replacing current video file buffer\n");
    ff_close_movie(ff);
  }

  // initialize values
  ff->pFrameFMT = NULL;
  ff->framerate = ff->duration = ff->frames = 1;
  ff->file_frame_offset = 0.0;
  ff->videoStream=-1;
  ff->tpf=1.0;
  ff->avprev=0;

  /* Open video file */
  if(av_open_input_file(&ff->pFormatCtx, file_name, NULL, 0, NULL)!=0) {
    fprintf( stderr, "Cannot open video file '%s'\n", file_name);
    return (-1);
  }

  /* Retrieve stream information */
  if(av_find_stream_info(ff->pFormatCtx)<0) {
    fprintf( stderr, "Cannot find stream information in file %s\n", file_name);
    av_close_input_file(ff->pFormatCtx);
    return (-1);
  }

  if (want_verbose) dump_format(ff->pFormatCtx, 0, file_name, 0);

  /* Find the first video stream */
  for(i=0; i<ff->pFormatCtx->nb_streams; i++)
#if LIBAVFORMAT_BUILD > 4629
    if(ff->pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO)
#else
    if(ff->pFormatCtx->streams[i]->codec.codec_type==CODEC_TYPE_VIDEO)
#endif
    {
      ff->videoStream=i;
      break;
    }

  if(ff->videoStream==-1) {
    fprintf( stderr, "Cannot find a video stream in file %s\n", file_name);
    av_close_input_file(ff->pFormatCtx);
    return (-1);
  }

  ff_set_framerate(ff);

#if defined(__BIG_ENDIAN__) && (__ppc__) 
// this cast is weird, but it works.. the bytes seem to be in 'correct' order, but the two
// 4byte-words are swapped. ?!
// I wonder how this behaves on a 64bit arch 
// - maybe it's bug in ffmpeg or all video files I tried had a bad header :D
  int64_t dur = (int64_t) (ff->pFormatCtx->duration);
  ff->duration = ( ((double) (((dur&0xffffffff)<<32)|((dur>>32)&0xffffffff))) / (double) AV_TIME_BASE );
#else
//ff->duration = (double) (((double) (ff->pFormatCtx->duration - ff->pFormatCtx->start_time))/ (double) AV_TIME_BASE);
  ff->duration = (double) (((double) (ff->pFormatCtx->duration))/ (double) AV_TIME_BASE);
#endif
  ff->frames = (long) (ff->framerate * ff->duration);
  ff->tpf = 1.0/(av_q2d(ff->pFormatCtx->streams[ff->videoStream]->time_base)*ff->framerate);
  ff->file_frame_offset =  ff->framerate*((double) ff->pFormatCtx->start_time/ (double) AV_TIME_BASE);

  if (want_verbose) {
    fprintf(stdout, "frame rate: %g\n", ff->framerate);
    fprintf(stdout, "length in seconds: %g\n", ff->duration);
    fprintf(stdout, "total frames: %ld\n", ff->frames);
    fprintf(stdout, "start offset: %.0f [frames]\n", ff->file_frame_offset);
  }

  // Get a pointer to the codec context for the video stream
#if LIBAVFORMAT_BUILD > 4629
  ff->pCodecCtx=ff->pFormatCtx->streams[ff->videoStream]->codec;
#else
  ff->pCodecCtx=&(ff->pFormatCtx->streams[ff->videoStream]->codec);
#endif

  // somewhere around LIBAVFORMAT_BUILD  4630 
#ifdef AVFMT_FLAG_GENPTS
  if (ff->want_genpts)
    ff->pFormatCtx->flags|=AVFMT_FLAG_GENPTS;
//  ff->pFormatCtx->flags|=AVFMT_FLAG_IGNIDX;
#endif

  // Find the decoder for the video stream
  pCodec=avcodec_find_decoder(ff->pCodecCtx->codec_id);
  if(pCodec==NULL) {
    fprintf( stderr, "Cannot find a codec for file: %s\n", file_name);
    av_close_input_file(ff->pFormatCtx);
    return(-1);
  }
    
  // Open codec
  if(avcodec_open(ff->pCodecCtx, pCodec)<0) {
    fprintf( stderr, "Cannot open the codec for file %s\n", file_name);
    av_close_input_file(ff->pFormatCtx);
    return(-1);
  }
  
  if (!(ff->pFrame=avcodec_alloc_frame())) {
    fprintf( stderr, "Cannot allocate video frame buffer\n");
    avcodec_close(ff->pCodecCtx);
    av_close_input_file(ff->pFormatCtx);
    return(-1);
  }

  if (!(ff->pFrameFMT=avcodec_alloc_frame())) {
    fprintf( stderr, "Cannot allocate display frame buffer\n");
    av_free(ff->pFrame);
    avcodec_close(ff->pCodecCtx);
    av_close_input_file(ff->pFormatCtx);
    return(-1);
  }

  ff->out_width=-1;
  ff->out_height=-1;
  ff_init_moviebuffer(ff);

  ff->current_file=strdup(file_name);
  return(0);
}

void reset_video_head(ffst *ff, AVPacket *packet) {
	int             frameFinished=0;
fprintf(stderr, "DEBUG: resetting decoder - seek/playhead rewind.\n");
#if LIBAVFORMAT_BUILD < 4617
	av_seek_frame(ff->pFormatCtx, ff->videoStream, 0);
#else
	av_seek_frame(ff->pFormatCtx, ff->videoStream, 0, AVSEEK_FLAG_BACKWARD);
#endif
	avcodec_flush_buffers(ff->pCodecCtx);

	while (!frameFinished) {
		av_read_frame(ff->pFormatCtx, packet);
		if(packet->stream_index==ff->videoStream) 
		avcodec_decode_video(ff->pCodecCtx, ff->pFrame, &frameFinished, packet->data, packet->size);
		if(packet->data) av_free_packet(packet);
	}
	ff->avprev=0;
}

// Note: set this high (>1000) if transport stopped and to a low value (<100) if transport is running.
#define MAX_CONT_FRAMES (500)

int my_seek_frame (ffst *ff, AVPacket *packet, int64_t timestamp) {
  AVStream *v_stream;
  int rv=1;
  int64_t mtsb = 0;
  int frameFinished; 
  int nolivelock = 0;
  static int ffdebug = 0;

  if (ff->videoStream < 0) return (0); 
  v_stream = ff->pFormatCtx->streams[ff->videoStream];

  if (ff->want_ignstart)  // timestamps in the file start counting at ..->start_time 
    timestamp+= (int64_t) rint(ff->framerate*((double)ff->pFormatCtx->start_time / (double)AV_TIME_BASE));

  // TODO: assert  0 < timestamp + ts_offset - (..->start_time)   < length   
	
#if LIBAVFORMAT_BUILD > 4629 // verify this version 
# if 1  // TODO -> -F <ratio>
  timestamp*=ff->tpf;
# else	// THIS is eqivalent - or even better at rounding but
	// does not work with -F <double> 
  timestamp=av_rescale_q(timestamp,c1_Q,v_stream->time_base); 
  timestamp=av_rescale_q(timestamp,c1_Q,v_stream->r_frame_rate); //< timestamp/=framerate; 
# endif
#endif

#if LIBAVFORMAT_BUILD < 4617
  rv= av_seek_frame(ff->pFormatCtx, ff->videoStream, timestamp / framerate * 1000000LL); 
#else
  if (ff->seekflags==SEEK_ANY) { 
    rv= av_seek_frame(ff->pFormatCtx, ff->videoStream, timestamp, AVSEEK_FLAG_ANY | AVSEEK_FLAG_BACKWARD) ;
    avcodec_flush_buffers(ff->pCodecCtx);
  } else if (ff->seekflags==SEEK_KEY) { 
    rv= av_seek_frame(ff->pFormatCtx, ff->videoStream, timestamp, AVSEEK_FLAG_BACKWARD) ;
    avcodec_flush_buffers(ff->pCodecCtx);
  } else /* SEEK_CONTINUOUS */ if (ff->avprev >= timestamp || ((ff->avprev + 32*ff->tpf) < timestamp) ) { 
    // NOTE: only seek if last-frame is less then 32 frames behind 
    // else read continuously until we get there :D
    // FIXME 32: use keyframes interval of video file or cmd-line-arg as threshold.
    // TODO: do we know if there is a keyframe inbetween now (my_avprev)
    // and the frame to seek to?? - if so rather seek to that frame than read until then.
    // and if no keyframe in between my_avprev and ts - prevent backwards seeks even if 
    // timestamp-my_avprev > threshold! - Oh well.

    // seek to keyframe *BEFORE* this frame
    rv= av_seek_frame(ff->pFormatCtx, ff->videoStream, timestamp, AVSEEK_FLAG_BACKWARD) ;
    avcodec_flush_buffers(ff->pCodecCtx);
#if 1
    ff->pCodecCtx->hurry_up = 1; // check if we're not there yet ;)
#endif
  } 
#endif 

  ff->avprev = timestamp;
  if (rv < 0) return (0); // seek failed.

read_frame:
  nolivelock++;
  if(av_read_frame(ff->pFormatCtx, packet)<0) {
    if (!want_quiet) printf("Reached movie end\n");
    return (0);
  }
#if LIBAVFORMAT_BUILD >=4616
  if (av_dup_packet(packet) < 0) {
    printf("can not allocate packet\n");
    goto read_frame;
  }
#endif
  if(packet->stream_index!=ff->videoStream) {
//  fprintf(stderr, "Not a video frame\n");
    if (packet->data) av_free_packet(packet);
    goto read_frame;
  }
  /* backwards compatible - no cont. seeking (seekmode ANY or KEY ; cmd-arg: -K, -k)
   * do we want a AVSEEK_FLAG_ANY + SEEK_CONTINUOUS option ?? not now.  */
#if LIBAVFORMAT_BUILD < 4617
  return (1);
#endif

  if (ff->seekflags!=SEEK_CONTINUOUS) return (1);

  mtsb = packet->pts;  
  if (mtsb == AV_NOPTS_VALUE) { 
    mtsb = packet->dts;
    if (ffdebug==0) { ffdebug=1; if (!want_quiet) fprintf(stderr,"WARNING: video file does not report pts information.\n         resorting to ffmpeg decompression timestamps.\n         consider to transcode the file or use the --genpts option.\n"); }
  }
  if (mtsb == AV_NOPTS_VALUE) { 
    if (ffdebug<2) { ffdebug=2; if (!want_quiet) fprintf(stderr,"ERROR: neither the video file nor the ffmpeg decoder were able to\n       provide a video frame timestamp."); }
    av_free_packet(packet);
    return (0);
  }

  if (mtsb >= timestamp) {
#if 1
    ff->pCodecCtx->hurry_up = 0; 
#endif
    return (1); // ok!
  }

  /* skip to next frame */

  avcodec_decode_video(ff->pCodecCtx, ff->pFrame, &frameFinished, packet->data, packet->size);
  av_free_packet(packet);
  if (!frameFinished) goto read_frame;
  if (nolivelock < MAX_CONT_FRAMES) goto read_frame;
  reset_video_head(ff, packet);
  return (0); // seek failed.
}

/**
 * seeks to frame and decodes and scales video frame
 *
 * @arg ptr handle / ff-data structure
 * @arg frame video frame to seek to
 */
void ff_render(void *ptr, unsigned long frame) {
  ffst *ff = (ffst*) ptr;
  int frameFinished;
  int64_t timestamp = (int64_t) frame;

  // if (ff->avprev == timestamp) return;
  //printf("Debug: ff-render %lu\n",frame);
 
  if (ff->pFrameFMT && ff->pFormatCtx && my_seek_frame(ff, &ff->packet, timestamp)) {
    while (1) { /* Decode video frame */
      frameFinished=0;	
      if(ff->packet.stream_index==ff->videoStream) 
	avcodec_decode_video(ff->pCodecCtx, ff->pFrame, &frameFinished, ff->packet.data, ff->packet.size);
      if(frameFinished) { /* Convert the image from its native format to FMT */
#if 1 // convert & scale
#ifdef HAVE_SWSCALE
        ff->pSWSCtx = sws_getCachedContext(ff->pSWSCtx, ff->pCodecCtx->width, ff->pCodecCtx->height, ff->pCodecCtx->pix_fmt, ff->out_width, ff->out_height, ff->render_fmt, SWS_BICUBIC, NULL, NULL, NULL);
        sws_scale(ff->pSWSCtx, ff->pFrame->data, ff->pFrame->linesize, 0, ff->pCodecCtx->height, ff->pFrameFMT->data, ff->pFrameFMT->linesize);
#else
	img_convert((AVPicture *)ff->pFrameFMT, ff->render_fmt, 
	    (AVPicture*)ff->pFrame, ff->pCodecCtx->pix_fmt,
	    ff->out_width, ff->out_height); // TODO: test this
	    //ff->pCodecCtx->width, ff->pCodecCtx->height);
#endif
#endif
	av_free_packet(&ff->packet); /* XXX */
	break;
      } else  { 
        fprintf( stderr, "Frame not finished..\n");
	if(ff->packet.data) av_free_packet(&ff->packet);
	if(av_read_frame(ff->pFormatCtx, &ff->packet)<0) { 
	  fprintf( stderr, "read error!\n");
	  reset_video_head(ff, &ff->packet);
	  //render_empty_frame(ff);
	  break;
	}
#if LIBAVFORMAT_BUILD >=4616
	if (av_dup_packet(&ff->packet) < 0) {
	  printf("can not allocate packet\n");
	  break;
	}
#endif
      }
    } /* end while !frame_finished */
  } else {
    if (ff->pFrameFMT && !want_quiet) fprintf( stderr, "frame seek unsucessful (frame: %lu).\n",frame);
    //render_empty_frame(ff);
  }
}


int main (int argc, char ** argv) {
  ff_initialize();
  ffst *ff = (ffst*) calloc(1,sizeof(ffst));
  ff->render_fmt=PIX_FMT_YUV420P;
  //ff->render_fmt=PIX_FMT_RGB24;
  ff->seekflags = SEEK_CONTINUOUS;
  ff->want_ignstart = 0;
  ff->want_genpts = 0;
  ff->packet.data = NULL;

  ff_open_movie(ff, argv[1]);

  unsigned long int frameno=0;
  while (frameno< ff->frames) {
    //printf("%lu\n", frameno);
    ff_render(ff,frameno++);
  }

  ff_close_movie(ff);
  free(ff);
  return (0);
}

/* vi:set ts=8 sts=2 sw=2: */
